/*!
  @file      SNP_DivFile.c
  @brief     分割データファイル入力ユーティリティ
  @date      March 2011
*/
#include "SNP_DivFile.h"
#include <stdlib.h>

/*!
  分割データファイル/ヘッダ情報構造体の初期化
  @param    phead     ヘッダ情報構造体へのポインタ
 */
void Init_SNP_DivData_Head(SNP_DivData_Head* phead) {
  if ( ! phead ) return;
  phead->blockNum = 0;
  phead->totalSNPNum = 0;
  phead->SNPStart = NULL;
  phead->SNPNum = NULL;
}

/*!
  分割データファイル/ヘッダ情報構造体のクリア
  @param    phead     ヘッダ情報構造体へのポインタ
 */
void Clear_SNP_DivData_Head(SNP_DivData_Head* phead) {
  if ( ! phead ) return;
  if ( phead->SNPStart ) free(phead->SNPStart);
  if ( phead->SNPNum ) free(phead->SNPNum);
  Init_SNP_DivData_Head(phead);
}

/*!
  分割データファイル/ヘッダ情報のファイル入力
  @param    fp        分割データファイルのファイルポインタ
  @param    phead     読み込み先ヘッダ情報構造体へのポインタ
 */
int Read_SNP_DivData_Head(FILE *fp, SNP_DivData_Head* phead) {
  int i;
  if ( ! fp ) return -1;
  if ( ! phead ) return -1;

  // rewind fp
  rewind(fp);

  // read blockNum, must be greater than 0
  if ( fread(&phead->blockNum, sizeof(long), 1, fp) != 1 )
    return -1;
  if ( phead->blockNum < 1 )
    return -1;

  // allocate memory for SNPStart and SNPNum
  phead->SNPStart = (long*)malloc(sizeof(long) * phead->blockNum);
  phead->SNPNum = (long*)malloc(sizeof(long) * phead->blockNum);
  if ( ! phead->SNPStart || ! phead->SNPNum ) {
    Clear_SNP_DivData_Head(phead);
    return -1;
  }

  // read SNPStart and SNPNum
  for ( i = 0; i < phead->blockNum; i++ ) {
    long buff[2];
    if ( fread(buff, sizeof(long), 2, fp) != 2 ) {
      Clear_SNP_DivData_Head(phead);
      return -1;
    }
    phead->SNPStart[i] = buff[0];
    phead->SNPNum[i] = buff[1];
  } // end of for(i)

  // read totalSNPNum
  if ( fread(&phead->totalSNPNum, sizeof(long), 1, fp) != 1 )
    return -1;

  return 0;
}

/*!
  分割データファイル/SNPデータのファイル入力
  @param    fp        分割データファイルのファイルポインタ
  @param    data      読み込み先のSNPデータ構造体
  @param    sampleNum 読み込むSNPデータのサンプル数
  @attention ファイルポインタは読み込むSNPデータ格納位置にある必要があります
 */
int Read_SNPdata(FILE *fp, SnpData* data, long sampleNum) {
  if ( ! fp || ! data ) return -1;
  if ( sampleNum < 0 ) return -1; // 0 is ok

  if ( fread(data->rsNumber, sizeof(char), RS_NUM_LEN, fp) != RS_NUM_LEN )
    return -1;
  if ( fread(data->SNPalleles, sizeof(char), SNP_NUM, fp) != SNP_NUM )
    return -1;
  if ( fread(data->chrom, sizeof(char), CHROM_LEN, fp) != CHROM_LEN )
    return -1;
  if ( fread(&data->pos, sizeof(long), 1, fp) != 1 )
    return -1;
  if ( fread(&data->allelesNum, sizeof(int), 1, fp) != 1 )
    return -1;
  if ( fread(&data->dataNum, sizeof(int), 1, fp) != 1 )
    return -1;
  if ( fread(data->SNPallelesCount, sizeof(int), SNP_COUNT, fp) != SNP_COUNT )
    return -1;
  if ( fread(&data->missingDataCount, sizeof(int), 1, fp) != 1 )
    return -1;
  if ( sampleNum > 0 ) {
    data->SNPdata = (char*)realloc(data->SNPdata, sizeof(char)*sampleNum);
    if ( ! data->SNPdata ) return -1;
    if ( fread(data->SNPdata, sizeof(char), sampleNum, fp) != sampleNum )
      return -1;
  }
  return 0;
}

/*!
  分割データファイルのファイルポインタを指定されたCASEデータの位置にseekする
  @param    fp        分割データファイルのファイルポインタ
  @param    caseIdx   seekするCASEデータのインデックス番号
 */
int Skip_To_CaseData(FILE *fp, long caseIdx)
{
  long rdBuff;
  long blockNum, caseSampleNum;

  if ( ! fp ) return -1;
  if ( caseIdx < 0 ) return -1;

  // rewind
  rewind(fp);

  // read blockNum of this file
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  blockNum = rdBuff;
  if ( blockNum < 1 ) return -1;

  // skip header (don't read totalSNPNum)
  if ( fseek(fp, sizeof(long)*(blockNum * 2 + 1), SEEK_CUR) != 0 )
    return -1;

  // read caseSampleNum
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  caseSampleNum = rdBuff;
  if ( caseSampleNum < 0 ) return -1; // 0 is ok

  // skip to caseIdx
  if ( caseIdx > 0 ) {
    if ( fseek(fp, (SNP_DATA_BASE_SIZE + caseSampleNum) * caseIdx, SEEK_CUR)
	 != 0 )
      return -1;
  }

  return 0;
}

/*!
  分割データファイルのファイルポインタを指定されたCONTROLデータの位置にseekする
  @param    fp         分割データファイルのファイルポインタ
  @param    controlIdx seekするCONTROLデータのインデックス番号
 */
int Skip_To_ControlData(FILE *fp, long controlIdx)
{
  long rdBuff;
  long blockNum, totalSNPNum, caseSampleNum, controlSampleNum;

  if ( ! fp ) return -1;
  if ( controlIdx < 0 ) return -1;

  // rewind
  rewind(fp);

  // read blockNum of this file
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  blockNum = rdBuff;
  if ( blockNum < 1 ) return -1;

  // skip header
  if ( fseek(fp, sizeof(long)*(blockNum * 2), SEEK_CUR) != 0 )
    return -1;

  // read totalSNPNum
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  totalSNPNum = rdBuff;
  if ( totalSNPNum < 1 ) return -1;

  // read caseSampleNum
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  caseSampleNum = rdBuff;
  if ( caseSampleNum < 0 ) return -1; // 0 is ok

  // skip case data
  if ( fseek(fp, (SNP_DATA_BASE_SIZE + caseSampleNum) * totalSNPNum, SEEK_CUR)
       != 0 )
    return -1;

  // read controlSampleNum
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  controlSampleNum = rdBuff;
  if ( controlSampleNum < 0 ) return -1; // 0 is ok

  // skip to controlIdx
  if ( controlIdx > 0 ) {
    if ( fseek(fp, (SNP_DATA_BASE_SIZE + controlSampleNum) * controlIdx,
	       SEEK_CUR) != 0 )
      return -1;
  }

  return 0;
}

/*!
  分割データファイルのファイルポインタを指定されたブロック領域データの位置にseekする
  @param    fp           分割データファイルのファイルポインタ
  @param    blockAreaIdx seekするブロック領域データのインデックス番号
 */
int Skip_To_BlockArea(FILE *fp, long blockAreaIdx)
{
  long rdBuff;
  long blockNum, totalSNPNum, caseSampleNum, controlSampleNum, blockAreaNum;

  if ( ! fp ) return -1;
  if ( blockAreaIdx < 0 ) return -1;

  // rewind
  rewind(fp);

  // read blockNum of this file
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  blockNum = rdBuff;
  if ( blockNum < 1 ) return -1;

  // skip header
  if ( fseek(fp, sizeof(long)*(blockNum * 2), SEEK_CUR) != 0 )
    return -1;

  // read totalSNPNum
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  totalSNPNum = rdBuff;
  if ( totalSNPNum < 1 ) return -1;

  // read caseSampleNum
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  caseSampleNum = rdBuff;
  if ( caseSampleNum < 0 ) return -1; // 0 is ok

  // skip case data
  if ( fseek(fp, (SNP_DATA_BASE_SIZE + caseSampleNum) * totalSNPNum, SEEK_CUR)
       != 0 )
    return -1;

  // read controlSampleNum
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  controlSampleNum = rdBuff;
  if ( controlSampleNum < 0 ) return -1; // 0 is ok

  // skip control data
  if ( fseek(fp, (SNP_DATA_BASE_SIZE + controlSampleNum) * totalSNPNum,
	     SEEK_CUR) != 0 )
    return -1;

  // read blockAreaNum
  if ( fread(&rdBuff, sizeof(long), 1, fp) != 1 )
    return -1;
  blockAreaNum = rdBuff;
  if ( blockAreaNum < 2 ) return -1; // blockAreaNum must be blockNum+1

  // skip to blockAreaIdx
  if ( blockAreaIdx > 0 ) {
    if ( fseek(fp, sizeof(long) * blockAreaIdx, SEEK_CUR) != 0 )
      return -1;
  }

  return 0;
}


/*!
  データ分割情報に基づき担当するブロック領域が格納されるファイルと、開始・終了ブロックを計算する
  @param    divFileSet     データ分割情報構造体へのポインタ
  @param    blockPartStart 担当するブロック領域のデータ全体における開始位置
  @param    blockPartNum   担当するブロック領域の総数
  @param    start_FileIdx  担当するSNPデータが格納された最初の分割データファイルのインデックス番号格納領域
  @param    end_FileIdx    担当するSNPデータが格納された最後の分割データファイルのインデックス番号格納領域
  @param    start_blockIdx 最初の分割データファイル中の最初の担当ブロック領域インデックス番号格納領域
  @param    end_blockIdx   最後の分割データファイル中の最後の担当ブロック領域インデックス番号格納領域
 */
int Check_BlockAreaRange(SNP_DivFileSet* divFileSet,
			 long blockPartStart, long blockPartNum,
			 int *start_FileIdx_, long *start_blockIdx_,
			 int *end_fileIdx_, long *end_blockIdx_)
{
  int i, j;
  int start_FileIdx; long start_blockIdx;
  int end_fileIdx; long end_blockIdx;
  int fileIdx, blockCnt, lastBlock;

  if ( ! divFileSet ) return -1;

  start_FileIdx = end_fileIdx = 0;
  start_blockIdx = end_blockIdx = 0;

  fileIdx = 0;
  while ( fileIdx < divFileSet->divNum ) {
    lastBlock = divFileSet->snpFileList[fileIdx].blockPartStart +
      divFileSet->snpFileList[fileIdx].blockPartNum;
    if ( blockPartStart < lastBlock ) break;
    fileIdx++;
  } // end of while
  if ( fileIdx >= divFileSet->divNum ) return -1; // out of range

  start_FileIdx = fileIdx;
  start_blockIdx
    = blockPartStart - divFileSet->snpFileList[start_FileIdx].blockPartStart;

  blockCnt
    = divFileSet->snpFileList[start_FileIdx].blockPartNum - start_blockIdx;
  if ( blockCnt >= blockPartNum ) {
    end_fileIdx = fileIdx;
    end_blockIdx = start_blockIdx + blockPartNum -1;
    goto _SUCCEED;
  }

  fileIdx++;
  while ( fileIdx < divFileSet->divNum ) {
    blockCnt += divFileSet->snpFileList[fileIdx].blockPartNum;
    if ( blockCnt >= blockPartNum ) break;
    fileIdx++;
  } // end of while
  if ( fileIdx >= divFileSet->divNum ) return -1; // out of range
  blockCnt -= divFileSet->snpFileList[fileIdx].blockPartNum;
  end_fileIdx = fileIdx;
  end_blockIdx = blockPartNum - blockCnt -1;
  

 _SUCCEED:
  *start_FileIdx_ = start_FileIdx;
  *start_blockIdx_ = start_blockIdx;
  *end_fileIdx_ = end_fileIdx;
  *end_blockIdx_ = end_blockIdx;
  return 0;
}


/*!
  データ分割情報に基づきブロック領域毎のデータ開始位置およびデータ数を設定する
  @param    divFileSet     データ分割情報構造体へのポインタ
  @param    start_FileIdx  担当するSNPデータが格納された最初の分割データファイルのインデックス番号
  @param    end_FileIdx    担当するSNPデータが格納された最後の分割データファイルのインデックス番号
  @param    start_blockIdx 最初の分割データファイル中の最初の担当ブロック領域インデックス番号
  @param    end_blockIdx   最後の分割データファイル中の最後の担当ブロック領域インデックス番号
  @param    linkSNPStart   ブロック領域毎のデータ開始位置格納領域へのポインタ
  @param    linkSNPNum     ブロック領域毎のデータ数格納領域へのポインタ
  @param    dataNum        担当する総SNPデータ数格納領域へのポインタ
  @attention linkSNPStartおよびlinkSNPNumは、担当する総ブロック領域数分確保されていなければなりません
 */
int Link_SNP_DivData(SNP_DivFileSet *divFileSet,
		     int start_FileIdx, int end_fileIdx,
		     long start_blockIdx, long end_blockIdx,
		     long *linkSNPStart, long *linkSNPNum, long *dataNum)
{
  int i, j, k, l;
  long startPos = 0;
  long endPos = 0;
  FILE *fpIn = NULL;
  SNP_DivData_Head snpDivDataHead;      

  if ( ! divFileSet || ! linkSNPStart || ! linkSNPNum || ! dataNum )
    return -1;

  Init_SNP_DivData_Head(&snpDivDataHead);
  *dataNum = 0;
  k = 0;
  for ( i = start_FileIdx; i <= end_fileIdx; i++ ) {
    char fileName[SNP_PATH_LEN*2];

    sprintf(fileName, "%s/%s",
	    divFileSet->baseDirName, divFileSet->snpFileList[i].path);
    fpIn = fopen(fileName, "rb");
    if ( ! fpIn ) {
      goto finalize;
    }

    if ( Read_SNP_DivData_Head(fpIn, &snpDivDataHead) != 0 ) {
      goto finalize;
    }

    if ( i == start_FileIdx )
      startPos = start_blockIdx;
    else
      startPos = 0;
    if ( i == end_fileIdx )
      endPos = end_blockIdx;
    else
      endPos = divFileSet->snpFileList[i].blockPartNum -1;

    for ( j = startPos; j <= endPos; j++ ) {
      linkSNPStart[k] = *dataNum;
      linkSNPNum[k] = snpDivDataHead.SNPNum[j];
      *dataNum += snpDivDataHead.SNPNum[j];
      k++;
    } // end of for(j)

    Clear_SNP_DivData_Head(&snpDivDataHead);
    fclose(fpIn);
    fpIn = NULL;
  } // end of for(i)

  if ( fpIn ) fclose(fpIn);
  return 0;

 finalize:
  Clear_SNP_DivData_Head(&snpDivDataHead);
  if ( fpIn ) fclose(fpIn);
  return -1;
}


/*!
  分割データファイル群からのデータ入力
  @param    divFileSet     データ分割情報構造体へのポインタ
  @param    start_FileIdx  担当するSNPデータが格納された最初の分割データファイルのインデックス番号
  @param    end_FileIdx    担当するSNPデータが格納された最後の分割データファイルのインデックス番号
  @param    start_blockIdx 最初の分割データファイル中の最初の担当ブロック領域インデックス番号
  @param    end_blockIdx   最後の分割データファイル中の最後の担当ブロック領域インデックス番号
  @param    snpData1　　   CASEデータ格納領域へのポインタ
  @param    snpData2       CONTROLデータ格納領域へのポインタ
  @param    blockArea      ブロック領域データ格納領域へのポインタ
  @attention snpData1およびsnpData2は担当する総SNPデータ数分、blockAreaは担当する総ブロック領域数+1個分確保されていなければなりません
 */
int Read_SNP_DivData(SNP_DivFileSet *divFileSet,
		     int start_FileIdx, int end_fileIdx,
		     long start_blockIdx, long end_blockIdx,
		     SnpData *snpData1, SnpData *snpData2, long *blockArea)
{
  int i, j, k, l;
  int index;
  long startPos = 0;
  long endPos = 0;
  FILE *fpIn = NULL;
  SNP_DivData_Head snpDivDataHead;      

  if ( ! divFileSet || ! snpData1 || ! snpData2 )
    return -1;
  if ( divFileSet->blockAreaType == 0 && ! blockArea )
    return -1;

  Init_SNP_DivData_Head(&snpDivDataHead);

  k = 0;
  index = 0;
  for ( i = start_FileIdx; i <= end_fileIdx; i++ ) {
    char fileName[SNP_PATH_LEN*2];
    long local_startDataIdx, local_DataNum;
    long local_startBlockIdx, local_BlockNum;

    sprintf(fileName, "%s/%s",
	    divFileSet->baseDirName, divFileSet->snpFileList[i].path);
    fpIn = fopen(fileName, "rb");
    if ( ! fpIn ) {
      goto finalize;
    }

    if ( Read_SNP_DivData_Head(fpIn, &snpDivDataHead) != 0 ) {
      goto finalize;
    }

    if ( i == start_FileIdx )
      startPos = start_blockIdx;
    else
      startPos = 0;
    if ( i == end_fileIdx )
      endPos = end_blockIdx;
    else
      endPos = divFileSet->snpFileList[i].blockPartNum -1;

    // このファイルのデータ読み込み開始番号
    local_startDataIdx = snpDivDataHead.SNPStart[startPos];
    // このファイルから読み込むデータ個数
    local_DataNum = 0;
    for ( j = startPos; j <= endPos; j++ ) {
      local_DataNum += snpDivDataHead.SNPNum[j];
    } // end of for(j)

    // このファイルのブロック領域読み込み開始番号
    local_startBlockIdx = startPos;
    // このファイルから読み込むブロック領域
    local_BlockNum = (endPos - startPos + 1) + 1;

    Clear_SNP_DivData_Head(&snpDivDataHead);

    // caseデータ読み込み
    if ( Skip_To_CaseData(fpIn, local_startDataIdx) != 0 ) {
      goto finalize;
    }
    for ( l = 0; l < local_DataNum; l++ ) {
      if ( Read_SNPdata(fpIn, &snpData1[index + l], divFileSet->caseSampleNum) != 0 ) {
	goto finalize;
      }
    } // end of for(l)

    // controlデータ読み込み
    if ( Skip_To_ControlData(fpIn, local_startDataIdx) != 0 ) {
      goto finalize;
    }
    for ( l = 0; l < local_DataNum; l++ ) {
      if ( Read_SNPdata(fpIn, &snpData2[index + l], divFileSet->controlSampleNum) != 0 ) {
	goto finalize;
      }
    } // end of for(l)

    // blockAreaデータ読み込み
    if ( divFileSet->blockAreaType == 0 ) {
      if ( Skip_To_BlockArea(fpIn, local_startBlockIdx) != 0 ) {
	goto finalize;
      }
      if ( fread(&blockArea[k], sizeof(long), local_BlockNum, fpIn)
	   != local_BlockNum ) {
	goto finalize;
      }
    }

    index += local_DataNum;
    k += (endPos - startPos + 1);

    fclose(fpIn);
    fpIn = NULL;
  } // end of for(i)

  if ( fpIn ) fclose(fpIn);
  return 0;

 finalize:
  Clear_SNP_DivData_Head(&snpDivDataHead);
  if ( fpIn ) fclose(fpIn);
  return -1;
}
