/*!
  @file      SNP_DivFileXML.c
  @brief     データ分割情報XMLファイル入力ユーティリティ
  @date      March 2011
*/
#include <libxml/tree.h>
#include <libgen.h>

#include "SNP_DivFileXML.h"

/*!
  分割データファイル情報構造体の初期化
  @param    pdata     分割データファイル情報構造体へのポインタ
 */
void Init_SNP_DivFile(SNP_DivFile* pdata) {
  if ( ! pdata ) return;
  pdata->blockPartStart = 0;
  pdata->blockPartNum = 0;
  pdata->path[0] = '\0';
}

/*!
  データ分割情報構造体の初期化
  @param    pdset     データ分割情報構造体へのポインタ
 */
void Init_SNP_DivFileSet(SNP_DivFileSet* pdset) {
  if ( ! pdset ) return;
  pdset->progName[0] = '\0';
  pdset->caseDataType = 0;
  pdset->caseSampleNum = 0;
  pdset->caseFileName[0] = '\0';
  pdset->controlDataType = 0;
  pdset->controlSampleNum = 0;
  pdset->controlFileName[0] = '\0';
  pdset->blockAreaType = 0;
  pdset->blockAreaFileName[0] = '\0';
  pdset->dataNum = 0;
  pdset->blockNum = 0;
  pdset->divNum = 0;
  pdset->baseDirName[0] = '\0';
  pdset->snpFileList = NULL;
}

/*!
  データ分割情報構造体の分割データファイル情報リストのメモリ確保
  @param    pdset     データ分割情報構造体へのポインタ
  @param    num       確保する分割データファイル情報の数
 */
int Alloc_SNP_DivFileSet(SNP_DivFileSet* pdset, long num) {
  if ( ! pdset || num < 0 ) return -1;
  if ( pdset->snpFileList ) {
    free(pdset->snpFileList);
    pdset->snpFileList = NULL;
    pdset->divNum = 0;
  }
  if ( num > 0 ) {
    int i;
    pdset->snpFileList = (SNP_DivFile*)malloc(sizeof(SNP_DivFile)*num);
    if ( ! pdset->snpFileList ) return -1;
    for ( i = 0; i < num; i++ )
      Init_SNP_DivFile(&pdset->snpFileList[i]);
    pdset->divNum = num;
  }
  return 0;
}

/*!
  データ分割情報構造体の分割データファイル情報リストのメモリ開放
  @param    pdset     データ分割情報構造体へのポインタ
 */
void Clear_SNP_DivFileSet(SNP_DivFileSet* pdset) {
  if ( ! pdset ) return;
  (void)Alloc_SNP_DivFileSet(pdset, 0);
}


/*!
  データ分割情報XMLファイルの<SNP_Files>ノードのXMLパース
  @param    xnp        <SNP_Files>ノードへのポインタ
  @param    divFileSet XMLパース結果を格納するデータ分割情報構造体へのポインタ
 */
int ParseSnpFilesNode(xmlNodePtr xnp, SNP_DivFileSet* divFileSet) {
  static xmlChar* xs;
  xmlNodePtr cur;
  xmlNodePtr bodyNode;
  int divNum;
  char baseDirName[SNP_PATH_LEN] = {'\0', };
  int cntSnpFile;

  if ( ! xnp ) return -1;
  if ( ! divFileSet ) return -1;

  // is 'SNP_Files' node?
  if ( strcmp((const char*)xnp->name, "SNP_Files") ) return -1;

  // 'divNum' prop
  divNum = 0;
  if ( xs ) xmlFree(xs);
  xs = xmlGetProp(xnp, (const xmlChar*)"divNum");
  if ( xs && strlen((const char*)xs) > 0 )
    divNum = atoi((const char*)xs);
  if ( divNum < 1 ) return -1;

  // Alloc divFileSet->snpFileList (and set divFileSet->divNum)
  if ( Alloc_SNP_DivFileSet(divFileSet, divNum) != 0 ) {
    return -1;
  }

  // 'baseDir' prop
  if ( xs ) xmlFree(xs);
  xs = xmlGetProp(xnp, (const xmlChar*)"baseDir");
  if ( xs && strlen((const char*)xs) > 0 )
    strcpy(baseDirName, (const char*)xs);
  else
    strcpy(baseDirName, ".");
  strcpy(divFileSet->baseDirName, baseDirName);

  // child node
  cntSnpFile = 0;
  cur = xnp->xmlChildrenNode;
  while ( cur ) {
    // skip comment
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }

    // 'SNP_File' node
    if ( !strcmp((const char*)cur->name, "SNP_File") ) {
      int blockPartStart = 0, blockPartNum = 0;
      if ( xs ) xmlFree(xs);
      xs = xmlGetProp(cur, (const xmlChar*)"blockPartStart");
      if ( xs && strlen((const char*)xs) > 0 )
	blockPartStart = atoi((const char*)xs);

      if ( xs ) xmlFree(xs);
      xs = xmlGetProp(cur, (const xmlChar*)"blockPartNum");
      if ( xs && strlen((const char*)xs) > 0 )
	blockPartNum = atoi((const char*)xs);

      bodyNode = cur->xmlChildrenNode;
      if ( ! bodyNode || bodyNode->type != XML_TEXT_NODE )
        goto _NEXT_XML_NODE;
      if ( ! bodyNode->content || strlen((const char*)bodyNode->content) < 1 )
        goto _NEXT_XML_NODE;
      strcpy(divFileSet->snpFileList[cntSnpFile].path,
	     (const char*)bodyNode->content);
      divFileSet->snpFileList[cntSnpFile].blockPartStart = blockPartStart;
      divFileSet->snpFileList[cntSnpFile].blockPartNum = blockPartNum;
      cntSnpFile++;
    }

  _NEXT_XML_NODE:
    cur = cur->next;
  } // end of while(cur)

  if ( cntSnpFile != divNum ) return -1;
  return 0;
}

/*!
  データ分割情報XMLファイルのXMLパース
  @param    fileName   パースするXMLファイルのパス名
  @param    divFileSet XMLパース結果を格納するデータ分割情報構造体へのポインタ
 */
int ParseXMLFile(char *fileName, SNP_DivFileSet* divFileSet) {
  static xmlChar* xs;
  xmlDocPtr doc;
  xmlNodePtr cur;
  xmlNodePtr bodyNode;
  char *xmlDirName, wkStr[256];

  int caseFlg = 0, controlFlg = 0, blockAreaFlg = 0;
  int blockNumFlg = 0, snpFlg = 0;

  if ( ! fileName || strlen(fileName) < 1 ) return -1;
  if ( ! divFileSet ) return -1;
  Init_SNP_DivFileSet(divFileSet);

  //---- open XML document
  doc = xmlParseFile(fileName);
  if ( ! doc ) return -1;

  //---- search 'ParaHaplo_DivData' node
  cur = xmlDocGetRootElement(doc);
  while ( cur ) {
    if ( ! strcmp((const char*)cur->name, "ParaHaplo_DivData") ) break;
    cur = cur->next;
  }
  if ( ! cur ) {
    xmlFreeDoc(doc); return -1;
  }

  //---- parse child node
  cur = cur->xmlChildrenNode;
  while ( cur ) {
    // skip comment
    if ( cur->type == XML_COMMENT_NODE ) {
      cur = cur->next;
      continue;
    }

    // 'program' node
    if ( !strcmp((const char*)cur->name, "program") ) {
      bodyNode = cur->xmlChildrenNode;
      if ( ! bodyNode || bodyNode->type != XML_TEXT_NODE )
        goto _NEXT_XML_NODE;
      if ( ! bodyNode->content || strlen((const char*)bodyNode->content) < 1 )
        goto _NEXT_XML_NODE;
      strcpy(divFileSet->progName, (const char*)bodyNode->content);
    }

    // 'caseData' node
    else if ( !strcmp((const char*)cur->name, "caseData") ) {
      int caseDataType = 0, caseSampleNum = 0;
      if ( xs ) xmlFree(xs);
      xs = xmlGetProp(cur, (const xmlChar*)"type");
      if ( xs && strlen((const char*)xs) > 0 )
	caseDataType = atoi((const char*)xs);

      if ( xs ) xmlFree(xs);
      xs = xmlGetProp(cur, (const xmlChar*)"sampleNum");
      if ( xs && strlen((const char*)xs) > 0 )
	caseSampleNum = atoi((const char*)xs);

      bodyNode = cur->xmlChildrenNode;
      if ( ! bodyNode || bodyNode->type != XML_TEXT_NODE )
        goto _NEXT_XML_NODE;
      if ( ! bodyNode->content || strlen((const char*)bodyNode->content) < 1 )
        goto _NEXT_XML_NODE;
      strcpy(divFileSet->caseFileName, (const char*)bodyNode->content);
      divFileSet->caseDataType = caseDataType;
      divFileSet->caseSampleNum = caseSampleNum;
      caseFlg = 1;
    }

    // 'controlData' node
    else if ( !strcmp((const char*)cur->name, "controlData") ) {
      int controlDataType = 0, controlSampleNum = 0;
      if ( xs ) xmlFree(xs);
      xs = xmlGetProp(cur, (const xmlChar*)"type");
      if ( xs && strlen((const char*)xs) > 0 )
	controlDataType = atoi((const char*)xs);

      if ( xs ) xmlFree(xs);
      xs = xmlGetProp(cur, (const xmlChar*)"sampleNum");
      if ( xs && strlen((const char*)xs) > 0 )
	controlSampleNum = atoi((const char*)xs);

      bodyNode = cur->xmlChildrenNode;
      if ( ! bodyNode || bodyNode->type != XML_TEXT_NODE )
        goto _NEXT_XML_NODE;
      if ( ! bodyNode->content || strlen((const char*)bodyNode->content) < 1 )
        goto _NEXT_XML_NODE;
      strcpy(divFileSet->controlFileName, (const char*)bodyNode->content);
      divFileSet->controlDataType = controlDataType;
      divFileSet->controlSampleNum = controlSampleNum;
      controlFlg = 1;
    }

    // 'blockAreaData' node
    else if ( !strcmp((const char*)cur->name, "blockAreaData") ) {
      int blockAreaType = 0;
      if ( xs ) xmlFree(xs);
      xs = xmlGetProp(cur, (const xmlChar*)"type");
      if ( xs && strlen((const char*)xs) > 0 )
	blockAreaType = atoi((const char*)xs);

      bodyNode = cur->xmlChildrenNode;
      if ( ! bodyNode || bodyNode->type != XML_TEXT_NODE )
        goto _NEXT_XML_NODE;
      if ( ! bodyNode->content || strlen((const char*)bodyNode->content) < 1 )
        goto _NEXT_XML_NODE;
      strcpy(divFileSet->blockAreaFileName, (const char*)bodyNode->content);
      divFileSet->blockAreaType = blockAreaType;
      blockAreaFlg = 1;
    }

    // 'dataNum' node
    else if ( !strcmp((const char*)cur->name, "dataNum") ) {
      bodyNode = cur->xmlChildrenNode;
      if ( ! bodyNode || bodyNode->type != XML_TEXT_NODE )
        goto _NEXT_XML_NODE;
      if ( ! bodyNode->content || strlen((const char*)bodyNode->content) < 1 )
        goto _NEXT_XML_NODE;
      divFileSet->dataNum = atoi((const char*)bodyNode->content);
    }

    // 'blockNum' node
    else if ( !strcmp((const char*)cur->name, "blockNum") ) {
      bodyNode = cur->xmlChildrenNode;
      if ( ! bodyNode || bodyNode->type != XML_TEXT_NODE )
        goto _NEXT_XML_NODE;
      if ( ! bodyNode->content || strlen((const char*)bodyNode->content) < 1 )
        goto _NEXT_XML_NODE;
      divFileSet->blockNum = atoi((const char*)bodyNode->content);
      blockNumFlg = 1;
    }

    // 'SNP_Files' node
    else if ( !strcmp((const char*)cur->name, "SNP_Files") ) {
      if ( ParseSnpFilesNode(cur, divFileSet) != 0 )
	goto _NEXT_XML_NODE;
      snpFlg = 1;
    }

  _NEXT_XML_NODE:
    cur = cur->next;
  } // end of while(cur)

  xmlFreeDoc(doc);

  // check flags
  if ( caseFlg * controlFlg * blockAreaFlg * blockNumFlg * snpFlg == 0 )
    return -1;

  // adjust baseDir: baseDir = baseDir(fileName) + "/" + baseDir
  strcpy(wkStr, fileName);
  xmlDirName = dirname(wkStr);
  if ( strcmp(divFileSet->baseDirName, ".") == 0 ) {
    strcpy(divFileSet->baseDirName, xmlDirName);
  } else {
    char bkupBaseDir[256];
    strcpy(bkupBaseDir, divFileSet->baseDirName);
    sprintf(divFileSet->baseDirName, "%s/%s", xmlDirName, bkupBaseDir);
  }

  return 0;
}

